﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Forms;
using System.IO;
using System.IO.Ports;
using System.Timers;
using System.Runtime.Serialization.Formatters.Binary;

namespace aFrameEdit
{
    public partial class MainForm : Form
    {
        private const int statusOK = 0;
        private const int statusNG = -1;

        private const int CategoryMain = 0;
        private const int CategorySub = 1;
        private const int CategoryXtra = 2;
        private const int CategoryDry = 3;
        private const int CategoryMute = 4;
        private const int CategoryMix = 5;
        private const int dsp1PatchNum = 80;
        private const int dsp2PatchNum = 80;
        private const int dspMemoryGrp = 40;

        private double delta = System.Math.Pow(2, 1.0 / 12.0);

        private System.Timers.Timer timer; // for serial transfer.
        public CurrentToneInfo currentToneInfo;
        private CurrentToneInfo currentToneInfoCopyForCancel;
        public CurrentToneData currentToneData;
        private CurrentToneData currentToneDataCopyForCancel;
        public ProjectToneNameList prjToneNameList;
        public ProjectToneNameList prjToneNameListCopy;
        public ProjectGroupList prjGroupList;
        private ToneName toneNameCopyForCancel;

        private bool isEditMode = false;
        public bool isTonePairChanged = false;
        public bool isToneInstEdited = false;
        public bool isToneEffectEdited = false;
        private int selectInstCateg = 0;
        private int selectInstCategPre = -1;

        private Label[] InstNameLabelArray;
        private ComboBox[] InstComboBoxArray;
        private Button[] InstDecButtonArray;
        private Button[] InstIncButtonArray;
        private Label[] InstUnitLabelArray;
        private Label[] EffectNameLabelArray;
        private ComboBox[] EffectComboBoxArray;
        private Button[] EffectDecButtonArray;
        private Button[] EffectIncButtonArray;
        private Label[] EffectUnitLabelArray;

        private Dictionary<int, double> instFreqTypePreVal;
        private Dictionary<int, double> effectFreqTypePreVal;

        /***** Form *****/
        public MainForm()
        {
            InitializeComponent();
            makeControlCollection();
            initializeToneInfo();
            InitComPort();
            currentToneInfo = new CurrentToneInfo();
            currentToneInfoCopyForCancel = new CurrentToneInfo();
            currentToneData = new CurrentToneData();
            currentToneDataCopyForCancel = new CurrentToneData();
            prjToneNameList = new ProjectToneNameList();
            prjToneNameListCopy = new ProjectToneNameList();
            prjGroupList = new ProjectGroupList();
            instFreqTypePreVal = new Dictionary<int, double>();
            effectFreqTypePreVal = new Dictionary<int, double>();
            makeFreqPrevalDictionary();
            timer = new System.Timers.Timer();
            timer.Elapsed += new ElapsedEventHandler(OnElapsedTimer);
            timer.Interval = 5000; // 5sec
            timer_con.Enabled = true;
            disableUI();
        }

        /***** UI control *****/
        private object GetControlArrayByName(Form frm, string name)
        {
            System.Collections.ArrayList ctrs =
                new System.Collections.ArrayList();
            object obj;
            for (int i = 0;
                (obj = FindControlByFieldName(frm, name + i.ToString())) != null;
                i++)
                ctrs.Add(obj);
            if (ctrs.Count == 0)
                return null;
            else
                return ctrs.ToArray(ctrs[0].GetType());
        }

        private static object FindControlByFieldName(Form frm, string name)
        {
            System.Type t = frm.GetType();

            System.Reflection.FieldInfo fi = t.GetField(
                name,
                System.Reflection.BindingFlags.Public |
                System.Reflection.BindingFlags.NonPublic |
                System.Reflection.BindingFlags.Instance |
                System.Reflection.BindingFlags.DeclaredOnly);

            if (fi == null)
                return null;

            return fi.GetValue(frm);
        }

        private void setLabels()
        {
            comboBox_inst_name.Text = currentToneInfo.toneName.instName;
            comboBox_effect_name.Text = currentToneInfo.toneName.effectName;
            string grpAndNum = getCurGrpAndNumString(currentToneInfo.mGrp, currentToneInfo.mNum);
            label_tone_info.Text = "GRP " + grpAndNum + "/" + currentToneInfo.mMax.ToString();
        }

        private void initializeToneInfo()
        {
            label_Version.Text = "VER.X.XX";
            comboBox_inst_name.Text = "XXXXX XXXXX";
            comboBox_effect_name.Text = "XXXXX XXXXX";
            label_tone_info.Text = "XXXXX XXXXX";

            for (int i = 0; i < 17; i++)
            {
                InstNameLabelArray[i].Text = "#" + i.ToString();
                InstNameLabelArray[i].Visible = true;
                InstComboBoxArray[i].DropDownStyle = ComboBoxStyle.Simple;
                InstComboBoxArray[i].Items.Clear();
                InstComboBoxArray[i].Text = "";
                InstComboBoxArray[i].Visible = true;
            }

            for (int i = 0; i < 16; i++)
            {
                EffectNameLabelArray[i].Text = "#" + i.ToString();
                EffectNameLabelArray[i].Visible = true;
                EffectComboBoxArray[i].DropDownStyle = ComboBoxStyle.Simple;
                EffectComboBoxArray[i].Items.Clear();
                EffectComboBoxArray[i].Text = "";
                EffectComboBoxArray[i].Visible = true;
            }
            selectInstCateg = 0;
            selectInstCategPre = -1;
        }

        private void enableUI()
        {
            button_group_original.Enabled = true;
            button_group_save.Enabled = true;
            button_group_saveAs.Enabled = true;
            button_bank_A.Enabled = true;
            button_bank_B.Enabled = true;
            button_bank_C.Enabled = true;
            button_bank_D.Enabled = true;
            button_FNC.Enabled = true;
            button_prev_tone.Enabled = true;
            button_next_tone.Enabled = true;
            label_tone_info.Enabled = true;

            button_inst_original.Enabled = true;
            button_inst_save.Enabled = true;
            button_inst_saveAs.Enabled = true;
            comboBox_inst_name.Enabled = true;
            button_categ_Main.Enabled = true;
            button_categ_Sub.Enabled = true;
            button_categ_Xtra.Enabled = true;
            button_categ_Dry.Enabled = true;
            button_categ_Mute.Enabled = true;
            button_categ_Mix.Enabled = true;

            for (int i = 0; i < 17; i++)
            {
                InstNameLabelArray[i].Enabled = true;
                InstComboBoxArray[i].Enabled = true;
                InstDecButtonArray[i].Enabled = true;
                InstIncButtonArray[i].Enabled = true;
                InstUnitLabelArray[i].Enabled = true;
            }

            button_effect_original.Enabled = true;
            button_effect_save.Enabled = true;
            button_effect_saveAs.Enabled = true;
            comboBox_effect_name.Enabled = true;
            for (int i = 0; i < 16; i++)
            {
                EffectNameLabelArray[i].Enabled = true;
                EffectComboBoxArray[i].Enabled = true;
                EffectDecButtonArray[i].Enabled = true;
                EffectIncButtonArray[i].Enabled = true;
                EffectUnitLabelArray[i].Enabled = true;
            }
        }

        private void disableUI()
        {
            button_group_original.Enabled = false;
            button_group_save.Enabled = false;
            button_group_saveAs.Enabled = false;
            button_bank_A.Enabled = false;
            button_bank_B.Enabled = false;
            button_bank_C.Enabled = false;
            button_bank_D.Enabled = false;
            button_FNC.Enabled = false;
            button_prev_tone.Enabled = false;
            button_next_tone.Enabled = false;
            label_tone_info.Enabled = false;

            button_inst_original.Enabled = false;
            button_inst_save.Enabled = false;
            button_inst_saveAs.Enabled = false;
            comboBox_inst_name.Enabled = false;
            button_categ_Main.Enabled = false;
            button_categ_Sub.Enabled = false;
            button_categ_Xtra.Enabled = false;
            button_categ_Dry.Enabled = false;
            button_categ_Mute.Enabled = false;
            button_categ_Mix.Enabled = false;

            for (int i = 0; i < 17; i++)
            {
                InstNameLabelArray[i].Enabled = false;
                InstComboBoxArray[i].Enabled = false;
                InstDecButtonArray[i].Enabled = false;
                InstIncButtonArray[i].Enabled = false;
                InstUnitLabelArray[i].Enabled = false;
            }

            button_effect_original.Enabled = false;
            button_effect_save.Enabled = false;
            button_effect_saveAs.Enabled = false;
            comboBox_effect_name.Enabled = false;
            for (int i = 0; i < 16; i++)
            {
                EffectNameLabelArray[i].Enabled = false;
                EffectComboBoxArray[i].Enabled = false;
                EffectDecButtonArray[i].Enabled = false;
                EffectIncButtonArray[i].Enabled = false;
                EffectUnitLabelArray[i].Enabled = false;
            }
        }

        public void setToneNameColor()
        {
            if (isTonePairChanged == true)
            {
                comboBox_inst_name.ForeColor = Color.Yellow;
                comboBox_effect_name.ForeColor = Color.Yellow;
            }
            else
            {
                comboBox_inst_name.ForeColor = Color.White;
                comboBox_effect_name.ForeColor = Color.White;
            }
        }

        private void setInstCategButton()
        {
            switch (selectInstCateg)
            {
                case 0: // "Main"
                    button_categ_Main.BackColor = Color.Red;
                    break;
                case 1: // "Sub"
                    button_categ_Sub.BackColor = Color.Red;
                    break;
                case 2: // "Xtra"
                    button_categ_Xtra.BackColor = Color.Red;
                    break;
                case 3: // "Dry"
                    button_categ_Dry.BackColor = Color.Red;
                    break;
                case 4: // "Mute"
                    button_categ_Mute.BackColor = Color.Red;
                    break;
                case 5: // "Mix"
                default:
                    button_categ_Mix.BackColor = Color.Red;
                    break;
            }

            switch (selectInstCategPre)
            {
                case 0: // "Main"
                    button_categ_Main.BackColor = Color.Transparent;
                    break;
                case 1: // "Sub"
                    button_categ_Sub.BackColor = Color.Transparent;
                    break;
                case 2: // "Xtra"
                    button_categ_Xtra.BackColor = Color.Transparent;
                    break;
                case 3: // "Dry"
                    button_categ_Dry.BackColor = Color.Transparent;
                    break;
                case 4: // "Mute"
                    button_categ_Mute.BackColor = Color.Transparent;
                    break;
                case 5: // "Mix"
                default:
                    button_categ_Mix.BackColor = Color.Transparent;
                    break;
            }
        }

        private void setInstCategButtonAllTransparent()
        {
            button_categ_Main.BackColor = Color.Transparent;
            button_categ_Sub.BackColor = Color.Transparent;
            button_categ_Xtra.BackColor = Color.Transparent;
            button_categ_Dry.BackColor = Color.Transparent;
            button_categ_Mute.BackColor = Color.Transparent;
            button_categ_Mix.BackColor = Color.Transparent;
        }

        private void setToneSelButton(int grp)
        {
            Color color = (grp > 3) ? Color.Gold : Color.Lime;
            switch (grp)
            {
                case 0: // "A"
                case 4: // "A'"
                    button_bank_A.BackColor = color;
                    button_bank_B.BackColor = Color.Transparent;
                    button_bank_C.BackColor = Color.Transparent;
                    button_bank_D.BackColor = Color.Transparent;
                    break;
                case 1: // "B"
                case 5: // "B'"
                    button_bank_A.BackColor = Color.Transparent;
                    button_bank_B.BackColor = color;
                    button_bank_C.BackColor = Color.Transparent;
                    button_bank_D.BackColor = Color.Transparent;
                    break;
                case 2: // "C"
                case 6: // "C'"
                    button_bank_A.BackColor = Color.Transparent;
                    button_bank_B.BackColor = Color.Transparent;
                    button_bank_C.BackColor = color;
                    button_bank_D.BackColor = Color.Transparent;
                    break;
                case 3: // "D"
                case 7: // "D'"
                default:
                    button_bank_A.BackColor = Color.Transparent;
                    button_bank_B.BackColor = Color.Transparent;
                    button_bank_C.BackColor = Color.Transparent;
                    button_bank_D.BackColor = color;
                    break;
            }
            button_FNC.BackColor = color;
        }

        private void setToneSelButtonAllTransparent()
        {
            button_bank_A.BackColor = Color.Transparent;
            button_bank_B.BackColor = Color.Transparent;
            button_bank_C.BackColor = Color.Transparent;
            button_bank_D.BackColor = Color.Transparent;
            button_FNC.BackColor = Color.Transparent;
        }

        private void toneDataChangedUIControl(int sel)
        {
            if (sel == 0) // Inst
            {
                isToneInstEdited = true;
                button_inst_original.Enabled = true;
                button_inst_save.Enabled = true;
                comboBox_inst_name.Enabled = false;
            }
            else if (sel == 1) // Effect
            {
                isToneEffectEdited = true;
                button_effect_original.Enabled = true;
                button_effect_save.Enabled = true;
                comboBox_effect_name.Enabled = false;
            }
            else // Both
            {
                isToneInstEdited = true;
                button_inst_original.Enabled = true;
                button_inst_save.Enabled = true;
                comboBox_inst_name.Enabled = false;
                isToneEffectEdited = true;
                button_effect_original.Enabled = true;
                button_effect_save.Enabled = true;
                comboBox_effect_name.Enabled = false;
            }
        }

        // Inst
        private void setInstToneInfo()
        {
            int categStartIdx = atdaChlabelCategIdx[selectInstCateg];
            int categSize = atdaChlabelCategSize[selectInstCateg];
            List<string> itemList = new List<string>();
            ComboBoxStyle cbxStyle;

            int index = 0;
            while (categSize > 0)
            {
                InstNameLabelArray[index].Text = atdaChLabel[categStartIdx + index];
                itemList.Clear();
                cbxStyle = getAtdaComboStyleAndSetItemList(categStartIdx + index, ref itemList);
                InstComboBoxArray[index].DropDownStyle = cbxStyle; // simple or DropDwonList
                if (cbxStyle == ComboBoxStyle.DropDownList)
                {
                    InstComboBoxArray[index].Items.Clear();
                    foreach (string item in itemList)
                    {
                        InstComboBoxArray[index].Items.Add(item);
                    }
                    InstComboBoxArray[index].SelectedIndex = currentToneData.instToneData.pVal[categStartIdx + index];
                    InstDecButtonArray[index].Visible = false;
                    InstIncButtonArray[index].Visible = false;
                }
                else
                {
                    InstComboBoxArray[index].Text = getAtdaValue(categStartIdx + index, currentToneData.instToneData.pVal[categStartIdx + index]);
                    InstDecButtonArray[index].Visible = true;
                    InstIncButtonArray[index].Visible = true;
                }
                bool isChanged = currentToneData.instToneData.isChanged[categStartIdx + index];
                if (isChanged == true)
                {
                    InstNameLabelArray[index].ForeColor = Color.Yellow;
                }
                else
                {
                    InstNameLabelArray[index].ForeColor = Color.White;
                }
                int valType = atdaValTypeTable[categStartIdx + index];
                int value = currentToneData.instToneData.pVal[categStartIdx + index];
                if (valType == ValTypeFreq)
                {
                    instFreqTypePreVal[categStartIdx + index] = (double)value;
                }
                InstNameLabelArray[index].Visible = true;
                InstComboBoxArray[index].Visible = true;
                string unitStr = atdaUnitInfo[categStartIdx + index];
                if (unitStr != "")
                {
                    InstUnitLabelArray[index].Text = unitStr;
                    InstUnitLabelArray[index].Visible = true;
                }
                else
                {
                    InstUnitLabelArray[index].Visible = false;
                }
                index += 1;
                categSize -= 1;
            }
            while (index < 17)
            {
                InstNameLabelArray[index].Visible = false;
                InstComboBoxArray[index].Visible = false;
                InstDecButtonArray[index].Visible = false;
                InstIncButtonArray[index].Visible = false;
                InstUnitLabelArray[index].Visible = false;
                index += 1;
            }
            label_tone_info.Focus();
        }

        private void enableInstToneUI()
        {
            for (int i = 0; i < 17; i++)
            {
                InstComboBoxArray[i].Enabled = true;
            }
        }

        private void disableInstToneUI()
        {
            for (int i = 0; i < 17; i++)
            {
                InstComboBoxArray[i].Enabled = false;
            }
        }

        public void setComboBoxInstName(string name)
        {
            comboBox_inst_name.Text = name;
        }

        public void setComboBoxInstNameItems()
        {
            comboBox_inst_name.Items.Clear();
            foreach (string name in prjToneNameList.instNames)
            {
                comboBox_inst_name.Items.Add(name);
            }
        }

        public void enableComboboxInstName()
        {
            comboBox_inst_name.Enabled = true;
        }

        public void disableInstOriginalAndSaveButton()
        {
            button_inst_original.Enabled = false;
            button_inst_save.Enabled = false;
        }

        public void setInstNameLabelWhite()
        {
            for (int i = 0; i < 17; i++)
            {
                InstNameLabelArray[i].ForeColor = Color.White;
            }
        }

        // Effect
        private void setEffectToneInfo()
        {
            int paramSize = currentToneData.effectToneData.pNum;
            int algoNum = currentToneData.effectToneData.algo;
            List<string> itemList = new List<string>();
            ComboBoxStyle cbxStyle;

            int index = 0;
            while (paramSize > 0)
            {
                EffectNameLabelArray[index].Text = getEffectNameString(algoNum, index);
                itemList.Clear();
                cbxStyle = getEffectComboStyleAndSetItemList(algoNum, index, ref itemList);
                EffectComboBoxArray[index].DropDownStyle = cbxStyle; // simple or DropDwonList
                if (cbxStyle == ComboBoxStyle.DropDownList)
                {
                    EffectComboBoxArray[index].Items.Clear();
                    foreach (string item in itemList)
                    {
                        EffectComboBoxArray[index].Items.Add(item);
                    }
                    EffectComboBoxArray[index].SelectedIndex = currentToneData.effectToneData.pVal[index];
                    EffectDecButtonArray[index].Visible = false;
                    EffectIncButtonArray[index].Visible = false;
                }
                else
                {
                    EffectComboBoxArray[index].Text = getEffectValString(algoNum, index, currentToneData.effectToneData.pVal[index]);
                    EffectDecButtonArray[index].Visible = true;
                    EffectIncButtonArray[index].Visible = true;
                }
                bool isChanged = currentToneData.effectToneData.isChanged[index];
                if (isChanged == true)
                {
                    EffectNameLabelArray[index].ForeColor = Color.Yellow;
                }
                else
                {
                    EffectNameLabelArray[index].ForeColor = Color.White;
                }
                int valType = getEffectValType(algoNum, index);
                int value = currentToneData.effectToneData.pVal[index];
                if (valType == ValTypeFreq)
                {
                    effectFreqTypePreVal[algoNum * 100 + index] = (double)value;
                }
                EffectNameLabelArray[index].Visible = true;
                EffectComboBoxArray[index].Visible = true;
                string unitStr = getEffectUnitInfoString(algoNum, index);
                if (unitStr != "")
                {
                    EffectUnitLabelArray[index].Text = unitStr;
                    EffectUnitLabelArray[index].Visible = true;
                }
                else
                {
                    EffectUnitLabelArray[index].Visible = false;
                }
                index += 1;
                paramSize -= 1;
            }
            while (index < 16)
            {
                EffectNameLabelArray[index].Visible = false;
                EffectComboBoxArray[index].Visible = false;
                EffectDecButtonArray[index].Visible = false;
                EffectIncButtonArray[index].Visible = false;
                EffectUnitLabelArray[index].Visible = false;
                index += 1;
            }
            label_tone_info.Focus();
        }

        private void enableEffectToneUI()
        {
            for (int i = 0; i < 16; i++)
            {
                EffectComboBoxArray[i].Enabled = true;
            }
        }

        private void disableEffectToneUI()
        {
            for (int i = 0; i < 16; i++)
            {
                EffectComboBoxArray[i].Enabled = false;
            }
        }

        public void setComboBoxEffectName(string name)
        {
            comboBox_effect_name.Text = name;
        }

        public void setComboBoxEffectNameItems()
        {
            comboBox_effect_name.Items.Clear();
            foreach (string name in prjToneNameList.effectNames)
            {
                comboBox_effect_name.Items.Add(name);
            }
        }

        public void enableComboBoxEffectName()
        {
            comboBox_effect_name.Enabled = true;
        }

        public void disableEffectOriginalAndSaveButton()
        {
            button_effect_original.Enabled = false;
            button_effect_save.Enabled = false;
        }

        public void setEffectNameLabelWhite()
        {
            for (int i = 0; i < 16; i++)
            {
                EffectNameLabelArray[i].ForeColor = Color.White;
            }
        }

        // Group
        public void disableGroupOriginalAndSaveButton()
        {
            button_group_original.Enabled = false;
            button_group_save.Enabled = false;
        }

        public void enableGroupOriginalAndSaveButton()
        {
            button_group_original.Enabled = true;
            button_group_save.Enabled = true;
        }

        public void setGroupOriginalButtonText(string name)
        {
            button_group_original.Text = name;
        }

        /***** Get Group function *****/
        private string getCurGrpAndNumString(int grp, int num)
        {
            string grpAndNumStr;

            if (grp < 4)
            {
                grpAndNumStr = (grp + 0xA).ToString("X") + "-" + (num + 1).ToString("00");
            }
            else
            {
                grpAndNumStr = ((grp & 0x3) + 0xA).ToString("X") + "'" + (num + 1).ToString("00");
            }
            return grpAndNumStr;
        }

        /***** Get Effect function *****/
        private string getEffectNameString(int algoNum, int index)
        {
            string retStr;

            switch (algoNum)
            {
                case 1: // Reverb
                    retStr = strevChLabel[index];
                    break;
                case 2: // Delay
                    retStr = stdelayChLabel[index];
                    break;
                case 3: // Chorus
                    retStr = stchorusChLabel[index];
                    break;
                case 4: // Flanger
                    retStr = stflangerChLabel[index];
                    break;
                case 5: // Phaser
                    retStr = stphaserChLabel[index];
                    break;
                case 6: // WahWah
                    retStr = stwahChLabel[index];
                    break;
                default:
                    retStr = "";
                    break;
            }
            return retStr;
        }

        private string getEffectValString(int algoNum, int index, int value)
        {
            string retStr;

            switch (algoNum)
            {
                case 1: // Reverb
                    retStr = getStrevValue(index, value);
                    break;
                case 2: // Delay
                    retStr = getStdelayValue(index, value);
                    break;
                case 3: // Chorus
                    retStr = getStchorusValue(index, value);
                    break;
                case 4: // Flanger
                    retStr = getStflangerValue(index, value);
                    break;
                case 5: // Phaser
                    retStr = getStphaserValue(index, value);
                    break;
                case 6: // WahWah
                    retStr = getStwahValue(index, value);
                    break;
                default:
                    retStr = "";
                    break;
            }
            return retStr;
        }

        private string getEffectMinValString(int algoNum, int index)
        {
            string retStr;

            switch (algoNum)
            {
                case 1: // Reverb
                    retStr = getStrevValue(index, strevMinValArray[index]);
                    break;
                case 2: // Delay
                    retStr = getStdelayValue(index, stdelayMinValArray[index]);
                    break;
                case 3: // Chorus
                    retStr = getStchorusValue(index, stchorusMinValArray[index]);
                    break;
                case 4: // Flanger
                    retStr = getStflangerValue(index, stflangerMinValArray[index]);
                    break;
                case 5: // Phaser
                    retStr = getStphaserValue(index, stphaserMinValArray[index]);
                    break;
                case 6: // WahWah
                    retStr = getStwahValue(index, stwahMinValArray[index]);
                    break;
                default:
                    retStr = "";
                    break;
            }
            return retStr;
        }

        private string getEffectMaxValString(int algoNum, int index)
        {
            string retStr;

            switch (algoNum)
            {
                case 1: // Reverb
                    retStr = getStrevValue(index, strevMaxValArray[index]);
                    break;
                case 2: // Delay
                    retStr = getStdelayValue(index, stdelayMaxValArray[index]);
                    break;
                case 3: // Chorus
                    retStr = getStchorusValue(index, stchorusMaxValArray[index]);
                    break;
                case 4: // Flanger
                    retStr = getStflangerValue(index, stflangerMaxValArray[index]);
                    break;
                case 5: // Phaser
                    retStr = getStphaserValue(index, stphaserMaxValArray[index]);
                    break;
                case 6: // WahWah
                    retStr = getStwahValue(index, stwahMaxValArray[index]);
                    break;
                default:
                    retStr = "";
                    break;
            }
            return retStr;
        }

        private int getEffectMinVal(int algoNum, int index)
        {
            int retInt;

            switch (algoNum)
            {
                case 1: // Reverb
                    retInt = strevMinValArray[index];
                    break;
                case 2: // Delay
                    retInt = stdelayMinValArray[index];
                    break;
                case 3: // Chorus
                    retInt = stchorusMinValArray[index];
                    break;
                case 4: // Flanger
                    retInt = stflangerMinValArray[index];
                    break;
                case 5: // Phaser
                    retInt = stphaserMinValArray[index];
                    break;
                case 6: // WahWah
                    retInt = stwahMinValArray[index];
                    break;
                default:
                    retInt = 0;
                    break;
            }
            return retInt;
        }

        private int getEffectMaxVal(int algoNum, int index)
        {
            int retInt;

            switch (algoNum)
            {
                case 1: // Reverb
                    retInt = strevMaxValArray[index];
                    break;
                case 2: // Delay
                    retInt = stdelayMaxValArray[index];
                    break;
                case 3: // Chorus
                    retInt = stchorusMaxValArray[index];
                    break;
                case 4: // Flanger
                    retInt = stflangerMaxValArray[index];
                    break;
                case 5: // Phaser
                    retInt = stphaserMaxValArray[index];
                    break;
                case 6: // WahWah
                    retInt = stwahMaxValArray[index];
                    break;
                default:
                    retInt = 0;
                    break;
            }
            return retInt;
        }

        private ComboBoxStyle getEffectComboStyleAndSetItemList(int algoNum, int index, ref List<string> itemList)
        {
            ComboBoxStyle cbxStyle;

            switch (algoNum)
            {
                case 1: // Reverb
                    cbxStyle = getRevComboStyleAndSetItemList(index, ref itemList);
                    break;
                case 2: // Delay
                    cbxStyle = getDelayComboStyleAndSetItemList(index, ref itemList);
                    break;
                case 3: // Chorus
                    cbxStyle = getChorusComboStyleAndSetItemList(index, ref itemList);
                    break;
                case 4: // Flanger
                    cbxStyle = getFlangerComboStyleAndSetItemList(index, ref itemList);
                    break;
                case 5: // Phaser
                    cbxStyle = getPhaserComboStyleAndSetItemList(index, ref itemList);
                    break;
                case 6: // WahWah
                    cbxStyle = getWahComboStyleAndSetItemList(index, ref itemList);
                    break;
                default:
                    cbxStyle = ComboBoxStyle.Simple;
                    break;
            }
            return cbxStyle;
        }

        private int getEffectParamValue(int algoNum, int index, string valStr, int preVal)
        {
            int retVal;

            switch (algoNum)
            {
                case 1: // Reverb
                    retVal = getStrevValue(index, valStr, preVal);
                    break;
                case 2: // Delay
                    retVal = getStdelayValue(index, valStr, preVal);
                    break;
                case 3: // Chorus
                    retVal = getStchrousValue(index, valStr, preVal);
                    break;
                case 4: // Flanger
                    retVal = getStflangerValue(index, valStr, preVal);
                    break;
                case 5: // Phaser
                    retVal = getStphaserValue(index, valStr, preVal);
                    break;
                case 6: // WahWah
                    retVal = getStwahValue(index, valStr, preVal);
                    break;
                default:
                    retVal = 0;
                    break;
            }
            return retVal;
        }

        private int getEffectValType(int algo, int index)
        {
            int retVal = 0;

            switch (algo)
            {
                case 1: // Reverb
                    retVal = strevValTypeTable[index];
                    break;
                case 2: // Delay
                    retVal = stdelayValTypeTable[index];
                    break;
                case 3: // Chorus
                    retVal = stchorusValTypeTable[index];
                    break;
                case 4: // Flanger
                    retVal = stflangerValTypeTable[index];
                    break;
                case 5: // Phaser
                    retVal = stphaserValTypeTable[index];
                    break;
                case 6: // WahWah
                    retVal = stwahValTypeTable[index];
                    break;
                default:
                    break;
            }
            return retVal;
        }

        private string getEffectUnitInfoString(int algo, int index)
        {
            string retStr = "";
            switch (algo)
            {
                case 1: // Reverb
                    retStr = strevUnitInfo[index];
                    break;
                case 2: // Delay
                    retStr = stdelayUnitInfo[index];
                    break;
                case 3: // Chorus
                    retStr = stchorusUnitInfo[index];
                    break;
                case 4: // Flanger
                    retStr = stflangerUnitInfo[index];
                    break;
                case 5: // Phaser
                    retStr = stphaserUnitInfo[index];
                    break;
                case 6: // WahWah
                    retStr = stwahUnitInfo[index];
                    break;
                default:
                    break;
            }
            return retStr;
        }

        /***** Deep Copy *****/
        private void copyCurrentToneInfo(ref CurrentToneInfo src, ref CurrentToneInfo dst)
        {
            dst = (CurrentToneInfo)objectUtil.DeepCopy(src);
        }

        private void copyCurrentToneData(ref CurrentToneData src, ref CurrentToneData dst)
        {
            dst = (CurrentToneData)objectUtil.DeepCopy(src);
        }

        private void copyProjectToneData(ref ProjectToneData src, ref ProjectToneData dst)
        {
            dst = (ProjectToneData)objectUtil.DeepCopy(src);
        }

        public void copyProjectToneNameList()
        {
            prjToneNameListCopy = (ProjectToneNameList)objectUtil.DeepCopy(prjToneNameList);
        }

        public void copyCurrentToneData()
        {
            currentToneDataCopyForCancel = (CurrentToneData)objectUtil.DeepCopy(currentToneData);
        }

        /***** API function *****/
        private void selectGroup()
        {
            int grp = currentToneInfo.mGrp;
            int num = currentToneInfo.mNum;
            string cmdExtSelectGroup = "aFE0:";
            List<string> aFrameReply = new List<string>();
            int status;

            writeData(cmdExtSelectGroup + grp.ToString() + "," + num.ToString() + ",1\r\n");
            status = waitAndGetLines(ref aFrameReply);
            if (status == statusOK)
            {
                if (int.Parse(aFrameReply[0]) != 0)
                {
#if DEBUG
                    MessageBox.Show("ExtSelectGroup() returns NG!!!");
#endif
                }
            }
            else // NG
            {
#if DEBUG
                MessageBox.Show("[ERR] selectGroup");
#endif
            }
        }

        private void selectGroup(int grp, int num)
        {
            string cmdExtSelectGroup = "aFE0:";
            List<string> aFrameReply = new List<string>();
            int status;

            writeData(cmdExtSelectGroup + grp.ToString() + "," + num.ToString() + ",1\r\n");
            status = waitAndGetLines(ref aFrameReply);
            if (status == statusOK)
            {
                if (int.Parse(aFrameReply[0]) != 0)
                {
#if DEBUG
                    MessageBox.Show("ExtSelectGroup() returns NG!!!");
#endif
                }
            }
            else // NG
            {
#if DEBUG
                MessageBox.Show("[ERR] selectGroup");
#endif
            }
        }

        private void writeGroup()
        {
            int grp = currentToneInfo.mGrp;
            int num = currentToneInfo.mNum;
            int max = currentToneInfo.mMax;
            string cmdExtWriteGroup = "aFE1:";
            List<string> aFrameReply = new List<string>();
            int status;

            writeData(cmdExtWriteGroup + grp.ToString() + "," + num.ToString() + "," + max.ToString() + ",1\r\n");
            status = waitAndGetLines(ref aFrameReply);
            if (status == statusOK)
            {
                if (int.Parse(aFrameReply[0]) != 0)
                {
#if DEBUG
                    MessageBox.Show("ExtWriteGroup() returns NG!!!");
#endif
                }
            }
            else // NG
            {
#if DEBUG
                MessageBox.Show("[ERR] writeGroup");
#endif
            }
        }

        public void writeGroup(int grp, int num, int max)
        {
            string cmdExtWriteGroup = "aFE1:";
            List<string> aFrameReply = new List<string>();
            int status;

            writeData(cmdExtWriteGroup + grp.ToString() + "," + num.ToString() + "," + max.ToString() + ",1\r\n");
            status = waitAndGetLines(ref aFrameReply);
            if (status == statusOK)
            {
                if (int.Parse(aFrameReply[0]) != 0)
                {
#if DEBUG
                    MessageBox.Show("ExtWriteGroup() returns NG!!!");
#endif
                }
            }
            else // NG
            {
#if DEBUG
                MessageBox.Show("[ERR] writeGroup");
#endif
            }
        }

        public int changeToneNum(int sel, int num)
        {
            List<string> aFrameReply = new List<string>();
            int status;
            string cmdExtChangeToneNum = "aFE2:";

            writeData(cmdExtChangeToneNum + sel.ToString() + "," + num.ToString() + ",1\r\n");
            status = waitAndGetLines(ref aFrameReply);
            if (status == statusOK)
            {
                if (int.Parse(aFrameReply[0]) != 0)
                {
#if DEBUG
                    MessageBox.Show("ExtChangeToneNum() returns NG!!!");
#endif
                    return statusNG;
                }
            }
            return statusOK;
        }

        private void changeEditBuffParam(int sel, int num, int value)
        {
            string cmdExtChangeEditBuffParam = "aFE3:";
            List<string> aFrameReply = new List<string>();
            int status;

            writeData(cmdExtChangeEditBuffParam + sel.ToString() + "," + num.ToString() + "," + value.ToString() + ",1\r\n");
            status = waitAndGetLines(ref aFrameReply);
            if (status == statusOK)
            {
                if (int.Parse(aFrameReply[0]) != 0)
                {
#if DEBUG
                    MessageBox.Show("ExtChangeEditBuffParam() returns NG!!!");
#endif
                }
            }
            else // NG
            {
#if DEBUG
                MessageBox.Show("[ERR] changeEditBuffParam");
#endif
            }
        }

        public void changeEditBuffName(int sel, string name)
        {
            string cmdExtChangeEditBuffName = "aFE4:";
            List<string> aFrameReply = new List<string>();
            int status;

            writeData(cmdExtChangeEditBuffName + sel.ToString() + ",1," + name + "\r\n");
            status = waitAndGetLines(ref aFrameReply);
            if (status == statusOK)
            {
                if (int.Parse(aFrameReply[0]) != 0)
                {
#if DEBUG
                    MessageBox.Show("ExtChangeEditBuffName() returns NG!!!");
#endif
                }
            }
            else // NG
            {
#if DEBUG
                MessageBox.Show("[ERR] changeEditBuffName");
#endif
            }
        }

        public void writeEditBuffToProject(int sel, int num)
        {
            string cmdExtWriteEditBuffToProject = "aFE5:";
            List<string> aFrameReply = new List<string>();
            int status;

            writeData(cmdExtWriteEditBuffToProject + sel.ToString() + "," + num.ToString() + ",1\r\n");
            status = waitAndGetLines(ref aFrameReply);
            if (status == statusOK)
            {
                if (int.Parse(aFrameReply[0]) != 0)
                {
#if DEBUG
                    MessageBox.Show("ExtWriteEditBuffToProject() returns NG!!!");
#endif
                }
            }
            else // NG
            {
#if DEBUG
                MessageBox.Show("[ERR] writeEditBuffToProject");
#endif
            }
        }

        private int getEditBuff(int sel)
        {          
            List<string> aFrameReply = new List<string>();
            int status;
            int retVal = statusOK;

            if (sel == 0) // Inst
            {
                writeData("aFE6:0,2,1\r\n");
                status = waitAndGetLines(5, ref aFrameReply);
                if (status == statusOK)
                {
                    if (checkCheckSum(aFrameReply) == statusOK)
                    {
                        currentToneData.instToneData.setProjectToneData(ref aFrameReply);
                    }
                    else
                    {
                        retVal = statusNG;
#if DEBUG
                        MessageBox.Show("[ERR] ExtGetEditBuff --- checkSum Error");
#endif
                    }
                }
                else // NG
                {
                    retVal = statusNG;
#if DEBUG
                    MessageBox.Show("[ERR] ExtGetEditBuff --- statusNG");
#endif
                }
            }
            else // Effect
            {
                writeData("aFE6:1,2,1\r\n");
                status = waitAndGetLines(5, ref aFrameReply);
                if (status == statusOK)
                {
                    if (checkCheckSum(aFrameReply) == statusOK)
                    {
                        currentToneData.effectToneData.setProjectToneData(ref aFrameReply);
                    }
                    else
                    {
                        retVal = statusNG;
#if DEBUG
                        MessageBox.Show("[ERR] ExtGetEditBuff --- checkSum Error");
#endif
                    }
                }
                else // NG
                {
                    retVal = statusNG;
#if DEBUG
                    MessageBox.Show("[ERR] ExtGetEditBuff --- statusNG");
#endif
                }
            }
            return retVal;
        }

        private void setEditBuff(int sel, string data)
        {
            List<string> aFrameReply = new List<string>();
            int status;

            if (sel == 0) // Inst
            {
                writeData("aFE7:0,2,1\r\n");
                status = waitAndGetLines(ref aFrameReply);
                if (status == statusOK)
                {
                    aFrameReply.Clear();
                    writeData(data);
                    status = waitAndGetLines(ref aFrameReply);
                    if (status == statusOK) // OK
                    {
                        switch (int.Parse(aFrameReply[0]))
                        {
                            case 1: // Timeout
#if DEBUG
                                MessageBox.Show("ExtSetEditBuff() returns TIME OUT.");
#endif
                                break;
                            case 2: // Data Size Error
#if DEBUG
                                MessageBox.Show("ExtSetEditBuff() returns DATA_SIZE_ERR");
#endif
                                break;
                            case 3: // CheckSum Error
#if DEBUG
                                MessageBox.Show("ExtSetEditBuff() returns CHECK_SUM_ERR");
#endif
                                break;
                            case 0: // OK
                            default:
                                break;
                        }
                    }
                    else // NG
                    {
#if DEBUG
                        MessageBox.Show("[ERR] ExtSetEditBuff <Data> --- statusNG");
#endif
                    }
                }
                else // NG
                {
#if DEBUG
                    MessageBox.Show("[ERR] ExtSetEditBuff <Header> --- statusNG");
#endif
                }
            }
            else // Effect
            {
                writeData("aFE7:1,2,1\r\n");
                status = waitAndGetLines(ref aFrameReply);
                if (status == statusOK)
                {
                    aFrameReply.Clear();
                    writeData(data);
                    status = waitAndGetLines(ref aFrameReply);
                    if (status == statusOK) // OK
                    {
                        switch (status)
                        {
                            case 1: // Timeout
#if DEBUG
                                MessageBox.Show("ExtSetEditBuff() returns TIME OUT.");
#endif
                                break;
                            case 2: // Data Size Error
#if DEBUG
                                MessageBox.Show("ExtSetEditBuff() returns DATA_SIZE_ERR");
#endif
                                break;
                            case 3: // CheckSum Error
#if DEBUG
                                MessageBox.Show("ExtSetEditBuff() returns CHECK_SUM_ERR");
#endif
                                break;
                            case 0: // OK
                            default:
                                break;
                        }
                    }
                    else
                    {
#if DEBUG
                        MessageBox.Show("[ERR] ExtSetEditBuff <Data> --- statusNG");
#endif
                    }
                }
                else // NG
                {
#if DEBUG
                        MessageBox.Show("[ERR] ExtSetEditBuff <Header> --- statusNG");
#endif
                }
            }
        }

        private void getCurrentToneInfo()
        {
            List<string> aFrameReply = new List<string>();
            int status;

            writeData("aFGA\r\n");
            status = waitAndGetLines(ref aFrameReply);
            if (status == statusOK)
            {
                currentToneInfo.setGroupInfo(ref aFrameReply);
            }
            else // NG
            {
#if DEBUG
                MessageBox.Show("[ERR] GetCurrentGroupToneNum");
#endif
            }

            aFrameReply.Clear();
            writeData("aFGB\r\n");
            status = waitAndGetLines(2, ref aFrameReply);
            if (status == statusOK)
            {
                currentToneInfo.setToneName(ref aFrameReply);
            }
            else // NG
            {
#if DEBUG
                MessageBox.Show("[ERR] GetCurrentToneName");
#endif
            }
        }

        public void getProjectGroupList(int group)
        {
            string cmdGetProjectGroupList = "aFGD:";
            List<string> aFrameReply = new List<string>();
            int status;

            writeData(cmdGetProjectGroupList + group.ToString() + "\r\n");
            status = waitAndGetLines(dspMemoryGrp + 1, ref aFrameReply);
            if (status == statusOK)
            {
                if (aFrameReply.Count == dspMemoryGrp + 1)
                {
                    prjGroupList.Clear();
                    foreach (string line in aFrameReply)
                    {
                        if (line.Split(':')[0] == "Gmax")
                        {
                            prjGroupList.gMax = int.Parse(line.Split(':')[1]);
                        }
                        else
                        {
                            prjGroupList.instNum.Add(int.Parse(line.Split(',')[0]));
                            prjGroupList.effectNum.Add(int.Parse(line.Split(',')[1]));
                        }
                    }
                }
                else
                {
#if DEBUG
                    MessageBox.Show("getProjectGroupList() --- List size is wrong.");
#endif
                }
            }
            else // NG
            {
#if DEBUG
                MessageBox.Show("[ERR] getProjectGroupList");
#endif
            }
        }

        private void getProjectToneNameList()
        {
            List<string> aFrameReply = new List<string>();
            int status;

            writeData("aFGE:0\r\n"); // Inst
            status = waitAndGetLines(dsp1PatchNum, ref aFrameReply);
            if (status == statusOK)
            {
                if (aFrameReply.Count == dsp1PatchNum)
                {
                    int idx = 1;
                    foreach (string line in aFrameReply)
                    {
                        prjToneNameList.instNames.Add("I" + idx.ToString("00") + ":" + line);
                        idx++;
                    }
                }
                else
                {
#if DEBUG
                    MessageBox.Show("GetProjectToneNameList() --- List size is wrong.");
#endif
                }
            }
            else // NG
            {
#if DEBUG
                MessageBox.Show("[ERR] getProjectToneNameList");
#endif
            }

            aFrameReply.Clear();
            writeData("aFGE:1\r\n"); // Effect
            status = waitAndGetLines(dsp2PatchNum, ref aFrameReply);
            if (status == statusOK)
            {
                if (aFrameReply.Count == dsp2PatchNum)
                {
                    int idx = 1;
                    foreach (string line in aFrameReply)
                    {
                        prjToneNameList.effectNames.Add("E" + idx.ToString("00") + ":" + line);
                        idx++;
                    }
                }
                else
                {
#if DEBUG
                    MessageBox.Show("GetProjectToneNameList() --- List size is wrong.");
#endif
                }
            }
            else // NG
            {
#if DEBUG
                MessageBox.Show("[ERR] getProjectToneNameList");
#endif
            }
        }

        /***** other *****/
        private int waitAndGetLines(int lineNum, ref List<string> replyData)
        {
            int status = 0;
            int rcvLineNum = 0;
            string rcvString = "";
            string[] rcvStringArray;

            while (rcvLineNum < lineNum)
            {
                isSerialTimerout = false;
                timer.Start();
                while (isSerialRcved == false)
                {
                    if (isSerialTimerout == true)
                        break;
                }
                timer.Stop();
                isSerialRcved = false;
                if (isSerialTimerout == false)
                {
                    string str = getStringfromRcvdData();
                    if (str != null)
                    {
                        rcvLineNum += str.Length - str.Replace("\n", "").Length;
                        rcvString += str;
                    }
                    else // null
                    {
                        status = -2;
                        break;
                    }
                }
                else // Timeout
                {
                    status = -1;
                    break;
                }
            }
            rcvStringArray = rcvString.Split(new string[] { "\r\n" }, StringSplitOptions.None);
            replyData.AddRange(rcvStringArray);
            replyData.RemoveAt(replyData.Count - 1); // Remove Last factor.
            return status;
        }

        private int waitAndGetLines(ref List<string> replyData)
        {
            int status = 0;

            isSerialTimerout = false;
            timer.Start();
            while (isSerialRcved == false)
            {
                if (isSerialTimerout == true)
                    break;
            }
            timer.Stop();
            isSerialRcved = false;
            if (isSerialTimerout == false)
            {
                string[] lines = getStringArrayfromRcvdData();
                if (lines != null)
                {
                    replyData.AddRange(lines);
                }
                else // null
                {
                    status = -2;
                }
            }
            else // Timeout
            {
                status = -1;
            }
            return status;
        }

        private int waitAndGetLines(ref string rcvString)
        {
            int status = 0;

            isSerialTimerout = false;
            timer.Start();
            while (isSerialRcved == false)
            {
                if (isSerialTimerout == true)
                    break;
            }
            timer.Stop();
            isSerialRcved = false;
            if (isSerialTimerout == false)
            {
                rcvString = getStringfromRcvdData();
            }
            else // Timeout
            {
                status = -1;
            }
            return status;
        }

        public void changeGroupInternal(int grp, int num)
        {
            selectGroup(grp, num);
            getCurrentToneInfo();
            int getStat = statusOK;
            for (int i = 0; i < 2; i++)
            {
                getStat |= getEditBuff(i);
            }
            currentToneData = (CurrentToneData)objectUtil.DeepCopy(currentToneDataCopyForCancel);
            if (getStat == statusOK)
            {
                setLabels();
                setToneSelButton(grp);
                setInstToneInfo();
                setEffectToneInfo();
            }
        }

        public void clearInstChangedInfo()
        {
            for (int i = 0; i < 67; i++)
            {
                currentToneData.instToneData.isChanged[i] = false;
                currentToneDataCopyForCancel.instToneData.isChanged[i] = false;
            }
        }

        public void clearEffectChangedInfo()
        {
            int num = currentToneData.effectToneData.pNum;
            for (int i = 0; i < num; i++)
            {
                currentToneData.effectToneData.isChanged[i] = false;
                currentToneDataCopyForCancel.effectToneData.isChanged[i] = false;
            }
        }

        private int alertTonePairChangeAndOperation()
        {
            int retVal = 0;

            if (isTonePairChanged == true)
            {
                DialogResult result = MessageBox.Show("Inst or/and Effect pair are changed.\nDo you want to overwrite?",
                    "Confirm", MessageBoxButtons.YesNoCancel, MessageBoxIcon.Exclamation, MessageBoxDefaultButton.Button3);

                if (result == DialogResult.Yes)
                {
                    writeGroup();
                    isTonePairChanged = false;
                }
                else if (result == DialogResult.No)
                {
                    changeToneNum(0, currentToneInfoCopyForCancel.instNum);
                    changeToneNum(1, currentToneInfoCopyForCancel.effectNum);
                    isTonePairChanged = false;
                }
                else if (result == DialogResult.Cancel)
                {
                    retVal = -1;
                }
            }
            setToneNameColor();
            return retVal;
        }

        private int alertToneChangeAndOperation()
        {
            int retVal = 0;

            if ((isToneInstEdited == true) || (isToneEffectEdited == true))
            {
                DialogResult result = MessageBox.Show("Inst or/and Effect Tone datas are changed.\nDoyou want to overwrite? ",
                    "Confirm", MessageBoxButtons.YesNoCancel, MessageBoxIcon.Exclamation, MessageBoxDefaultButton.Button3);

                if (result == DialogResult.Yes)
                {
                    writeEditBuffToProject(0, currentToneInfo.instNum);
                    writeEditBuffToProject(1, currentToneInfo.effectNum);
                    isToneInstEdited = false;
                    isToneEffectEdited = false;
                    button_inst_original.Enabled = false;
                    button_effect_original.Enabled = false;
                    button_inst_save.Enabled = false;
                    button_effect_save.Enabled = false;
                    comboBox_inst_name.Enabled = true;
                    comboBox_effect_name.Enabled = true;
                }
                else if (result == DialogResult.No)
                {
                    isToneInstEdited = false;
                    isToneEffectEdited = false;
                    button_inst_original.Enabled = false;
                    button_effect_original.Enabled = false;
                    button_inst_save.Enabled = false;
                    button_effect_save.Enabled = false;
                }
                else if (result == DialogResult.Cancel)
                {
                    retVal = -1;
                }
            }
            return retVal;
        }

        private int checkCheckSum(List<string> lines)
        {
            int checkSum = 0;

            int algo = int.Parse(lines[0]);
            string name = lines[1];
            int pNum = int.Parse(lines[2]);
            string[] pValStr = lines[3].Split(',');

            int CSum = int.Parse(lines[4]);

            // Algo
            byte[] bytes = BitConverter.GetBytes((Int16)algo);
            foreach (byte b in bytes)
            {
                checkSum += b;
            }

            // Name
            bytes = System.Text.Encoding.ASCII.GetBytes(name);
            foreach (byte b in bytes)
            {
                checkSum += b;
            }

            // PNum
            bytes = BitConverter.GetBytes((Int16)pNum);
            foreach (byte b in bytes)
            {
                checkSum += b;
            }

            // pVal
            foreach (string str in pValStr)
            {
                if (str != "")
                {
                    bytes = BitConverter.GetBytes(Int16.Parse(str));
                    foreach (byte b in bytes)
                    {
                        checkSum += b;
                    }
                }
            }

            if (checkSum == CSum)
            {
                return statusOK;
            }
            else
            {
                return statusNG;
            }
        }

        private void makeFreqPrevalDictionary()
        {
            instFreqTypePreVal.Add(3, 0.0);     // Main Tune
            instFreqTypePreVal.Add(15, 0.0);    // Sub Tune
            instFreqTypePreVal.Add(28, 0.0);    // Xtra Tune
            instFreqTypePreVal.Add(38, 0.0);    // DryC EqF
            instFreqTypePreVal.Add(41, 0.0);    // DryE EqF
            instFreqTypePreVal.Add(44, 0.0);    // Center LPF
            instFreqTypePreVal.Add(45, 0.0);    // Edge HPF
            effectFreqTypePreVal.Add(304, 0.0); // Wet HPF(chorus)
            effectFreqTypePreVal.Add(305, 0.0); // Wet LPF(chorus)
            effectFreqTypePreVal.Add(602, 0.0); // Freq Min(wah)
            effectFreqTypePreVal.Add(603, 0.0); // Freq Max(wah)
        }

        private void makeControlCollection()
        {
            InstNameLabelArray = (Label[])GetControlArrayByName(this, "label_inst_name_");
            InstComboBoxArray = (ComboBox[])GetControlArrayByName(this, "comboBox_inst_");
            InstDecButtonArray = (Button[])GetControlArrayByName(this, "button_Inst_dec_");
            InstIncButtonArray = (Button[])GetControlArrayByName(this, "button_Inst_inc_");
            InstUnitLabelArray = (Label[])GetControlArrayByName(this, "label_inst_unit_");
            EffectNameLabelArray = (Label[])GetControlArrayByName(this, "label_effect_name_");
            EffectComboBoxArray = (ComboBox[])GetControlArrayByName(this, "comboBox_effect_");
            EffectDecButtonArray = (Button[])GetControlArrayByName(this, "button_Effect_dec_");
            EffectIncButtonArray = (Button[])GetControlArrayByName(this, "button_Effect_inc_");
            EffectUnitLabelArray = (Label[])GetControlArrayByName(this, "label_effect_unit_");
        }

        private void makeToneList()
        {
            foreach (string name in prjToneNameList.instNames)
            {
                comboBox_inst_name.Items.Add(name);
            }
            foreach (string name in prjToneNameList.effectNames)
            {
                comboBox_effect_name.Items.Add(name);
            }
        }

        /**************************************************/
        /*                 Event handler                  */
        /**************************************************/
        /***** Click *****/
        private void button_group_save_Click(object sender, EventArgs e)
        {
            writeGroup();
            isTonePairChanged = false;
            button_group_save.Enabled = false;
            button_group_original.Enabled = false;
            button_group_original.Text = "Original";
            button_group_saveAs.Enabled = true;
            setToneNameColor();
        }

        private void button_group_original_Click(object sender, EventArgs e)
        {
            CurrentToneInfo tmpToneInfo = new CurrentToneInfo();
            copyCurrentToneInfo(ref currentToneInfo, ref tmpToneInfo);
            copyCurrentToneInfo(ref currentToneInfoCopyForCancel, ref currentToneInfo);
            copyCurrentToneInfo(ref tmpToneInfo, ref currentToneInfoCopyForCancel);
            int status = changeToneNum(0, currentToneInfo.instNum);
            if (status == 0)
            {
                if (getEditBuff(0) == statusOK)
                {
                    setInstToneInfo();
                }
                comboBox_inst_name.Text = prjToneNameList.instNames[currentToneInfo.instNum];
            }
            status = changeToneNum(1, currentToneInfo.effectNum);
            if (status == 0)
            {
                if (getEditBuff(1) == statusOK)
                {
                    setEffectToneInfo();
                }
                comboBox_effect_name.Text = prjToneNameList.effectNames[currentToneInfo.effectNum];
            }
            if (isTonePairChanged == true)
            {
                isTonePairChanged = false;
                button_group_original.Text = "Edited";
                button_group_saveAs.Enabled = false;
            }
            else
            {
                isTonePairChanged = true;
                button_group_original.Text = "Original";
                button_group_saveAs.Enabled = true;
            }
            setToneNameColor();
        }

        private void button_group_saveAs_Click(object sender, EventArgs e)
        {
            getProjectGroupList(currentToneInfo.mGrp);
            GroupSaveAsForm groupSaveAsForm = new GroupSaveAsForm(currentToneInfo.mGrp, prjGroupList, prjToneNameList);
            groupSaveAsForm.Owner = this;
            groupSaveAsForm.Show();
        }

        private void button_Inst_Click(object sender, EventArgs e)
        {
            string[] strArray = ((Button)sender).Name.Split('_');
            int index = int.Parse(strArray[3]);
            string dir = strArray[2];
            int categStartIdx = atdaChlabelCategIdx[selectInstCateg];
            int valType = atdaValTypeTable[categStartIdx + index];
            int preVal, value;
            if (valType == ValTypeFreq)
            {
                double preValD = instFreqTypePreVal[categStartIdx + index];
                double valueD;
                if (dir == "dec")
                {
                    valueD = preValD / delta;
                }
                else
                {
                    valueD = preValD * delta;
                }
                instFreqTypePreVal[categStartIdx + index] = valueD;
                value = (int)(valueD + 0.5);
            }
            else if (valType == ValTypePowTbl)
            {
                preVal = currentToneData.instToneData.pVal[categStartIdx + index];
                int idxVal = getPowTblIndex(preVal);
                if (dir == "dec")
                {
                    idxVal -= 1;
                }
                else
                {
                    idxVal += 1;
                }
                int absVal = System.Math.Abs(idxVal);
                value = MixPowTable[absVal];
                if (idxVal < 0)
                {
                    value *= -1;
                }
            }
            else
            {
                preVal = currentToneData.instToneData.pVal[categStartIdx + index];
                if (dir == "dec")
                {
                    value = preVal - 1;
                }
                else
                {
                    value = preVal + 1;
                }
            }
            int min = atdaMinValArray[categStartIdx + index];
            int max = atdaMaxValArray[categStartIdx + index];
            value = limitParam(min, max, value);
            currentToneData.instToneData.pVal[categStartIdx + index] = value;
            InstComboBoxArray[index].Text = getAtdaValue(categStartIdx + index, value);
            InstNameLabelArray[index].ForeColor = Color.Yellow;
            currentToneData.instToneData.isChanged[categStartIdx + index] = true;
            changeEditBuffParam(0, categStartIdx + index, value);
            toneDataChangedUIControl(0);
        }

        private void button_inst_save_Click(object sender, EventArgs e)
        {
            if (isToneInstEdited == true)
            {
                writeEditBuffToProject(0, currentToneInfo.instNum);
                isToneInstEdited = false;
                copyProjectToneData(ref currentToneData.instToneData, ref currentToneDataCopyForCancel.instToneData);
            }
            else
            {
                copyProjectToneData(ref currentToneData.instToneData, ref currentToneDataCopyForCancel.instToneData);
                enableInstToneUI();
                button_inst_original.Text = "Original";
                button_inst_saveAs.Enabled = true;
            }
            disableInstOriginalAndSaveButton();
            enableComboboxInstName();
            setInstNameLabelWhite();
            clearInstChangedInfo();
        }

        private void button_inst_original_Click(object sender, EventArgs e)
        {
            string checkSumString = currentToneDataCopyForCancel.instToneData.calcCheckSum().ToString() + "\r\n";
            setEditBuff(0, currentToneDataCopyForCancel.instToneData.getStringData() + checkSumString);
            ProjectToneData tmpToneData = new ProjectToneData();
            copyProjectToneData(ref currentToneData.instToneData, ref tmpToneData);
            copyProjectToneData(ref currentToneDataCopyForCancel.instToneData, ref currentToneData.instToneData);
            copyProjectToneData(ref tmpToneData, ref currentToneDataCopyForCancel.instToneData);
            setInstToneInfo();
            string tmpString = comboBox_inst_name.Text;
            comboBox_inst_name.Text = toneNameCopyForCancel.instName;
            toneNameCopyForCancel.instName = tmpString;
            changeEditBuffName(0, toneNameCopyForCancel.instName.Split(':')[1]);

            if (isToneInstEdited == true)
            {
                disableInstToneUI();
                button_inst_original.Text = "Edited";
                button_inst_saveAs.Enabled = false;
                isToneInstEdited = false;
            }
            else
            {
                enableInstToneUI();
                button_inst_original.Text = "Original";
                button_inst_saveAs.Enabled = true;
                isToneInstEdited = true;
            }
        }

        private void button_inst_saveAs_Click(object sender, EventArgs e)
        {
            InstToneSaveAsForm instSaveAsForm = new InstToneSaveAsForm(prjToneNameList, currentToneInfo.instNum);
            instSaveAsForm.Owner = this;
            instSaveAsForm.Show();
        }

        private void button_Effect_Click(object sender, EventArgs e)
        {
            string[] strArray = ((Button)sender).Name.Split('_');
            int index = int.Parse(strArray[3]);
            string dir = strArray[2];
            int algo = currentToneData.effectToneData.algo;
            int valType = getEffectValType(algo, index);
            int preVal, value;
            if (valType == ValTypeFreq)
            {
                double preValD = effectFreqTypePreVal[algo * 100 + index];
                double valueD;
                if (dir == "dec")
                {
                    valueD = preValD / delta;
                }
                else
                {
                    valueD = preValD * delta;
                }
                effectFreqTypePreVal[algo * 100 + index] = valueD;
                value = (int)(valueD + 0.5);
            }
            else if (valType == ValTypePowTbl)
            {
                preVal = currentToneData.effectToneData.pVal[index];
                int idxVal = getPowTblIndex(preVal);
                if (dir == "dec")
                {
                    idxVal -= 1;
                }
                else
                {
                    idxVal += 1;
                }
                int absVal = System.Math.Abs(idxVal);
                value = MixPowTable[absVal];
                if (idxVal < 0)
                {
                    value *= -1;
                }
            }
            else
            {
                preVal = currentToneData.effectToneData.pVal[index];
                if (dir == "dec")
                {
                    value = preVal - 1;
                }
                else
                {
                    value = preVal + 1;
                }
            }
            int min = getEffectMinVal(algo, index);
            int max = getEffectMaxVal(algo, index);
            value = limitParam(min, max, value);
            currentToneData.effectToneData.pVal[index] = value;
            EffectComboBoxArray[index].Text = getEffectValString(algo, index, value);
            EffectNameLabelArray[index].ForeColor = Color.Yellow;
            currentToneData.effectToneData.isChanged[index] = true;
            changeEditBuffParam(1, index, value);
            toneDataChangedUIControl(1);
        }

        private void button_effect_save_Click(object sender, EventArgs e)
        {
            if (isToneEffectEdited == true)
            {
                writeEditBuffToProject(1, currentToneInfo.effectNum);
                isToneEffectEdited = false;
                copyProjectToneData(ref currentToneData.effectToneData, ref currentToneDataCopyForCancel.effectToneData);
            }
            else
            {
                copyProjectToneData(ref currentToneData.effectToneData, ref currentToneDataCopyForCancel.effectToneData);
                enableEffectToneUI();
                button_effect_original.Text = "Original";
                button_effect_saveAs.Enabled = true;
            }
            disableEffectOriginalAndSaveButton();
            enableComboBoxEffectName();
            setEffectNameLabelWhite();
            clearEffectChangedInfo();
        }

        private void button_effect_original_Click(object sender, EventArgs e)
        {
            string checkSumString = currentToneDataCopyForCancel.effectToneData.calcCheckSum().ToString() + "\r\n";
            setEditBuff(1, currentToneDataCopyForCancel.effectToneData.getStringData() + checkSumString);
            ProjectToneData tmpToneData = new ProjectToneData();
            copyProjectToneData(ref currentToneData.effectToneData, ref tmpToneData);
            copyProjectToneData(ref currentToneDataCopyForCancel.effectToneData, ref currentToneData.effectToneData);
            copyProjectToneData(ref tmpToneData, ref currentToneDataCopyForCancel.effectToneData);
            setEffectToneInfo();
            string tmpString = comboBox_effect_name.Text;
            comboBox_effect_name.Text = toneNameCopyForCancel.effectName;
            toneNameCopyForCancel.effectName = tmpString;
            changeEditBuffName(1, toneNameCopyForCancel.effectName.Split(':')[1]);

            if (isToneEffectEdited == true)
            {
                disableEffectToneUI();
                button_effect_original.Text = "Edited";
                button_effect_saveAs.Enabled = false;
                isToneEffectEdited = false;
            }
            else
            {
                enableEffectToneUI();
                button_effect_original.Text = "Original";
                button_effect_saveAs.Enabled = true;
                isToneEffectEdited = true;
            }
        }

        private void button_effect_saveAs_Click(object sender, EventArgs e)
        {
            EffectToneSaveAsForm effectSaveAsForm = new EffectToneSaveAsForm(prjToneNameList, currentToneInfo.effectNum);
            effectSaveAsForm.Owner = this;
            effectSaveAsForm.Show();
        }

        private void button_tone_change_Click(object sender, EventArgs e)
        {
            if (alertToneChangeAndOperation() == -1)
            {
                return;
            }
            if (alertTonePairChangeAndOperation() == -1)
            {
                return;
            }

            string[] names = ((Button)sender).Name.Split('_'); // button_bank_A..D, button_FNC, button_prev_tone, button_next_tone

            if (names[1] == "prev")
            {
                int value = currentToneInfo.mNum - 1;
                if (value < 0)
                {
                    value = currentToneInfo.mMax - 1;
                }
                currentToneInfo.mNum = value;
            }
            else if (names[1] == "next")
            {
                int value = currentToneInfo.mNum + 1;
                if (value >= currentToneInfo.mMax)
                {
                    value = 0;
                }
                currentToneInfo.mNum = value;
            }
            else if (names[1] == "FNC")
            {
                int grp = currentToneInfo.mGrp;
                if (grp >= 4)
                {
                    grp -= 4;
                }
                else
                {
                    grp += 4;
                }
                currentToneInfo.mGrp = grp;
                currentToneInfo.mNum = 0;
            }
            else // bank
            {
                string bankName = names[2];
                setGrpAndNumWithBankName(bankName);
                
            }
            selectGroup();
            setToneSelButton(currentToneInfo.mGrp);
            getCurrentToneInfo();
            currentToneInfoCopyForCancel = (CurrentToneInfo)objectUtil.DeepCopy(currentToneInfo);
            toneNameCopyForCancel = currentToneInfo.toneName;
            int getStat = statusOK;
            for (int i = 0; i < 2; i++)
            {
                getStat |= getEditBuff(i);
            }
            copyCurrentToneData(ref currentToneData, ref currentToneDataCopyForCancel);
            if (getStat == statusOK)
            {
                setLabels();
                setInstToneInfo();
                setEffectToneInfo();
            }
            button_inst_original.Enabled = false;
            button_effect_original.Enabled = false;
        }

        private void button_categ_Click(object sender, EventArgs e)
        {
            string categName = ((Button)sender).Name.Split('_')[2];

            switch (categName)
            {
                case "Main":
                    if (selectInstCateg != CategoryMain)
                    {
                        selectInstCategPre = selectInstCateg;
                        selectInstCateg = CategoryMain;
                        setInstCategButton();
                        setInstToneInfo();
                    }
                    break;
                case "Sub":
                    if (selectInstCateg != CategorySub)
                    {
                        selectInstCategPre = selectInstCateg;
                        selectInstCateg = CategorySub;
                        setInstCategButton();
                        setInstToneInfo();
                    }
                    break;
                case "Xtra":
                    if (selectInstCateg != CategoryXtra)
                    {
                        selectInstCategPre = selectInstCateg;
                        selectInstCateg = CategoryXtra;
                        setInstCategButton();
                        setInstToneInfo();
                    }
                    break;
                case "Dry":
                    if (selectInstCateg != CategoryDry)
                    {
                        selectInstCategPre = selectInstCateg;
                        selectInstCateg = CategoryDry;
                        setInstCategButton();
                        setInstToneInfo();
                    }
                    break;
                case "Mute":
                    if (selectInstCateg != CategoryMute)
                    {
                        selectInstCategPre = selectInstCateg;
                        selectInstCateg = CategoryMute;
                        setInstCategButton();
                        setInstToneInfo();
                    }
                    break;
                case "Mix":
                    if (selectInstCateg != CategoryMix)
                    {
                        selectInstCategPre = selectInstCateg;
                        selectInstCateg = CategoryMix;
                        setInstCategButton();
                        setInstToneInfo();
                    }
                    break;      
                default:
                    break;
            }
        }  

        private void toolStripButton_Con_Click(object sender, EventArgs e)
        {
            List<string> aFrameReply = new List<string>();
            int status;

            connectComPort();
            writeData("aFG0\r\n"); // Get Version
            status = waitAndGetLines(ref aFrameReply);
            if (status == statusOK)
            {
                label_Version.Text = aFrameReply[0].Split('-')[0]; // Extract version number.

                aFrameReply.Clear();
                writeData("aFG1\r\n"); // Get Mode
                status = waitAndGetLines(ref aFrameReply);
                if (status == statusOK)
                {
                    int mode = int.Parse(aFrameReply[0]);
                    bool isConnect = false;
                    switch (mode)
                    {
                        case 0:
                            isConnect = true;
                            break;
                        case 9:
                            aFrameReply.Clear();
                            writeData("aFS1:0\r\n"); // Set "Ext Edit Mode" to "Mode Home".
                            status = waitAndGetLines(ref aFrameReply);
                            if ((status == statusOK) && (int.Parse(aFrameReply[0]) == 0))
                            {
                                isConnect = true;
                            }
                            break;
                        default:
                            MessageBox.Show("Change aFrame to HOME MODE");
                            break;
                    }
                    if (isConnect == true)
                    {
                        aFrameReply.Clear();
                        writeData("aFS1:1\r\n"); // Set "Mode Home" to "Ext Edit Mode".
                        status = waitAndGetLines(ref aFrameReply);
                        if (status == statusOK)
                        {
                            if (int.Parse(aFrameReply[0]) == 0)
                            {
                                isEditMode = true;
                            }
                            else
                            {
#if DEBUG
                                MessageBox.Show("SetExtModeSw() returns NG!!!");
#endif
                                return;
                            }
                        }
                        else
                        {
#if DEBUG
                            MessageBox.Show("SetExtModeSw --- statusNG");
#endif
                            return;
                        }
                    }
                    else
                    {
                        return;
                    }
                }
                else
                {
#if DEBUG
                    MessageBox.Show("GetMode --- statusNG");
#endif
                }
            }
            else
            {
                MessageBox.Show("It is not supported\r\nin this version");
                return;
            }

            getProjectToneNameList();
            makeToneList();
            getCurrentToneInfo();
            currentToneInfoCopyForCancel = (CurrentToneInfo)objectUtil.DeepCopy(currentToneInfo);
            toneNameCopyForCancel = currentToneInfo.toneName;
            setLabels();
            setToneSelButton(currentToneInfo.mGrp);
            setInstCategButton();
            int getStat = statusOK;
            for (int i = 0; i < 2; i++)
            {
                getStat |= getEditBuff(i);
            }
            copyCurrentToneData(ref currentToneData, ref currentToneDataCopyForCancel);
            enableUI();
            if (getStat == statusOK)
            {
                setInstToneInfo();
                setEffectToneInfo();
            }
            toolStripButton_Discon.Enabled = true;
            toolStripButton_Con.Enabled = false;
            button_group_original.Enabled = false;
            button_group_save.Enabled = false;
            isToneInstEdited = false;
            isToneEffectEdited = false;
            button_inst_original.Enabled = false;
            button_inst_save.Enabled = false;
            button_effect_original.Enabled = false;
            button_effect_save.Enabled = false;
            button_inst_original.Text = "Original";
            button_effect_original.Text = "Original";
        }

        private void toolStripButton_Discon_Click(object sender, EventArgs e)
        {
            List<string> aFrameReply = new List<string>();
            int status;

            if (alertToneChangeAndOperation() == -1)
            {
                return;
            }
            if (alertTonePairChangeAndOperation() == -1)
            {
                return;
            }

            if (isEditMode == true)
            {
                writeData("aFS1:0\r\n"); // Set "Ext Edit Mode" to "Mode Home".
                status = waitAndGetLines(ref aFrameReply);
                if (status == statusOK)
                {
                    if (int.Parse(aFrameReply[0]) != 0)
                    {
#if DEBUG
                        MessageBox.Show("SetExtModeSw() returns NG!!!");
#endif
                    }
                    else
                    {
                        isEditMode = false;
                    }
                }
            }
            disconnectComPort();
            setInstCategButtonAllTransparent();
            setToneSelButtonAllTransparent();
            disableUI();

            prjToneNameList.Clear();
            toolStripButton_Discon.Enabled = false;
            toolStripButton_Con.Enabled = true;
        }

        /***** SelectionChangeCommitted *****/
        private void comboBox_inst_SelectionChangeCommitted(object sender, EventArgs e)
        {
            int index = int.Parse(((ComboBox)sender).Name.Split('_')[2]);
            string text;

            if (((ComboBox)sender).DropDownStyle == ComboBoxStyle.Simple)
            {
                text = (string)((ComboBox)sender).SelectedText;
            }
            else // DropDownList
            {
                text = (string)((ComboBox)sender).SelectedItem;
            }
            int categStartIdx = atdaChlabelCategIdx[selectInstCateg];
            int valType = atdaValTypeTable[categStartIdx + index];
            int value = getAtdaValue(categStartIdx + index, text, currentToneData.instToneData.pVal[categStartIdx + index]);
            int min = atdaMinValArray[categStartIdx + index];
            int max = atdaMaxValArray[categStartIdx + index];
            value = limitParam(min, max, value);
            currentToneData.instToneData.pVal[categStartIdx + index] = value;
            changeEditBuffParam(0, categStartIdx + index, value); // Update Edit Buff
            if (valType == ValTypeFreq)
            {
                instFreqTypePreVal[categStartIdx + index] = (double)value;
            }
            ((ComboBox)sender).Text = getAtdaValue(categStartIdx + index, value);
            InstNameLabelArray[index].ForeColor = Color.Yellow;
            currentToneData.instToneData.isChanged[categStartIdx + index] = true;
            toneDataChangedUIControl(0);
        }

        private void comboBox_effect_SelectionChangeCommitted(object sender, EventArgs e)
        {
            int index = int.Parse(((ComboBox)sender).Name.Split('_')[2]);
            string text;

            if (((ComboBox)sender).DropDownStyle == ComboBoxStyle.Simple)
            {
                text = (string)((ComboBox)sender).SelectedText;
            }
            else // DropDownList
            {
                text = (string)((ComboBox)sender).SelectedItem;
            }
            int algo = currentToneData.effectToneData.algo;
            int valType = getEffectValType(algo, index);
            int value = getEffectParamValue(algo, index, text, currentToneData.effectToneData.pVal[index]);
            int min = getEffectMinVal(algo, index);
            int max = getEffectMaxVal(algo, index);
            value = limitParam(min, max, value);
            currentToneData.effectToneData.pVal[index] = value;
            changeEditBuffParam(1, index, value); // Update Edit Buff
            if (valType == ValTypeFreq)
            {
                effectFreqTypePreVal[algo * 100 + index] = (double)value;
            }
            ((ComboBox)sender).Text = getEffectValString(algo, index, value);
            EffectNameLabelArray[index].ForeColor = Color.Yellow;
            currentToneData.effectToneData.isChanged[index] = true;
            toneDataChangedUIControl(1);
        }

        private void comboBox_inst_name_SelectionChangeCommitted(object sender, EventArgs e)
        {
            int index = ((ComboBox)sender).SelectedIndex;
            string text = (string)((ComboBox)sender).SelectedItem;
            comboBox_inst_name.Text = text;
            currentToneInfo.instNum = index;
            currentToneInfo.toneName.instName = text;
            changeToneNum(0, index);
            int getStat = getEditBuff(0);
            copyProjectToneData(ref currentToneData.instToneData, ref currentToneDataCopyForCancel.instToneData);
            setLabels();
            toneNameCopyForCancel.instName = currentToneInfo.toneName.instName;
            if (getStat == statusOK)
            {
                setInstToneInfo();
            }
            isTonePairChanged = true;
            button_group_original.Enabled = true;
            button_group_save.Enabled = true;
            setToneNameColor();
        }

        private void comboBox_effect_name_SelectionChangeCommitted(object sender, EventArgs e)
        {
            int index = ((ComboBox)sender).SelectedIndex;
            string text = (string)((ComboBox)sender).SelectedItem;
            comboBox_effect_name.Text = text;
            currentToneInfo.effectNum = index;
            currentToneInfo.toneName.effectName = text;
            changeToneNum(1, index);
            int getStat = getEditBuff(1);
            copyProjectToneData(ref currentToneData.effectToneData, ref currentToneDataCopyForCancel.effectToneData);
            setLabels();
            toneNameCopyForCancel.effectName = currentToneInfo.toneName.effectName;
            if (getStat == statusOK)
            {
                setEffectToneInfo();
            }
            isTonePairChanged = true;
            button_group_original.Enabled = true;
            button_group_save.Enabled = true;
            setToneNameColor();
        }

        /***** Timer event *****/
        private void OnElapsedTimer(Object sender, ElapsedEventArgs e) // Timer Event Hnadler
        {
#if DEBUG
            MessageBox.Show("[ERR] Sorry, Timeout Occured!! ");
#endif
            isSerialTimerout = true;
        }

        private void timer_con_Tick(object sender, EventArgs e)
        {
            if (isDevFound == false)
            {
                InitComPort();
            }
        }

        /***** FormClosed *****/
        private void Form1_FormClosed(object sender, FormClosedEventArgs e)
        {
            List<string> aFrameReply = new List<string>();
            int status;

            if (isEditMode == true)
            {
                writeData("aFS1:0\r\n"); // Set "Ext Edit Mode" to "Mode Home".
                status = waitAndGetLines(ref aFrameReply);
                if (status == statusOK)
                {
                    if (int.Parse(aFrameReply[0]) != 0)
                    {
#if DEBUG
                        MessageBox.Show("SetExtModeSw() returns NG!!!");
#endif
                    }
                    else
                    {
                        isEditMode = false;
                    }
                }
            }
            disconnectComPort();
        }
    }

    /******* Other class define ***************/

    [Serializable]
    public class ToneName
    {
        public string instName = "";
        public string effectName = "";
    }

    [Serializable]
    public class CurrentToneInfo
    {
        public int mGrp = 0;
        public int mNum = 0;
        public int mMax = 0;
        public int instNum = 0;
        public int effectNum = 0;
        public ToneName toneName;

        public CurrentToneInfo()
        {
            toneName = new ToneName();
        }

        public void setGroupInfo(ref List<string> replyList)
        {
            string[] curToneNum = replyList[0].Split(',');
            if (curToneNum.Length == 5)
            {
                mGrp = int.Parse(curToneNum[0]);
                mNum = int.Parse(curToneNum[1]);
                mMax = int.Parse(curToneNum[2]);
                instNum = int.Parse(curToneNum[3]);
                effectNum = int.Parse(curToneNum[4]);
            }
        }

        public void setToneName(ref List<string> replyList)
        {
            toneName.instName = replyList[0];
            toneName.effectName = replyList[1];
        }
    }

    [Serializable]
    public class CurrentToneData
    {
        public ProjectToneData instToneData;
        public ProjectToneData effectToneData;

        public CurrentToneData()
        {
            instToneData = new ProjectToneData();
            effectToneData = new ProjectToneData();
        }
    }

    [Serializable]
    public class ProjectToneData
    {
        public int algo = 0;
        public string name = "";
        public int pNum = 0; // Number of Parameter.
        public List<int> pVal;
        public List<bool> isChanged;

        public ProjectToneData()
        {
            pVal = new List<int>();
            isChanged = new List<bool>();
        }

        public void setProjectToneData(ref List<string> replyList)
        {
            if (replyList.Count >= 4)
            {
                int paramNum = 0;
                algo = int.Parse(replyList[0]);    // Algo
                name = replyList[1];               // Name
                paramNum = int.Parse(replyList[2]);// PNum
                pNum = paramNum;

                string[] paramStr = replyList[3].Split(',');
                List<string> paramList = new List<string>();
                paramList.AddRange(paramStr);
                paramList.RemoveAt(paramList.Count - 1); // Last factor is "". and Remove.

                if (paramList.Count == paramNum)
                {
                    pVal.Clear();
                    isChanged.Clear();
                    foreach (string valStr in paramList)
                    {
                        pVal.Add(int.Parse(valStr));
                        isChanged.Add(false);
                    }
                }
                else
                {
#if DEBUG
                    MessageBox.Show("[ERR] getProjectToneData --- param number error");
#endif
                }
            }
            else
            {
#if DEBUG
                MessageBox.Show("[ERR] getProjectToneData --- received lines is less than expected lines.");
#endif
            }
        }

        public string getStringData()
        {
            string retString = "";

            // algo
            retString += algo.ToString() + "\r\n";

            // name
            retString += name + "\r\n";

            // pNum
            retString += pNum.ToString() + "\r\n";

            // pVal

            foreach (int val in pVal)
            {
                retString += val.ToString() + ",";
            }
            retString += "\r\n";

            return retString;
        }

        public int calcCheckSum()
        {
            int checkSum = 0;

            // Algo
            byte[] bytes = BitConverter.GetBytes((Int16)algo);
            foreach (byte b in bytes)
            {
                checkSum += b;
            }

            // Name
            bytes = System.Text.Encoding.ASCII.GetBytes(name);
            foreach (byte b in bytes)
            {
                checkSum += b;
            }

            // PNum
            bytes = BitConverter.GetBytes((Int16)pNum);
            foreach (byte b in bytes)
            {
                checkSum += b;
            }

            // pVal
            int[] pValArray = pVal.ToArray();
            foreach (int intVal in pValArray)
            {
                bytes = BitConverter.GetBytes((Int16)intVal);
                foreach (byte b in bytes)
                {
                    checkSum += b;
                }
            }

            return checkSum;
        }
    }

    public class ProjectGroupList
    {
        public int gMax = 0;
        public List<int> instNum;
        public List<int> effectNum;

        public ProjectGroupList()
        {
            instNum = new List<int>();
            effectNum = new List<int>();
        }

        public void Clear()
        {
            instNum.Clear();
            effectNum.Clear();
        }
    }

    [Serializable]
    public class ProjectToneNameList
    {
        public List<string> instNames;
        public List<string> effectNames;

        public ProjectToneNameList()
        {
            instNames = new List<string>();
            effectNames = new List<string>();
        }

        public void Clear()
        {
            instNames.Clear();
            effectNames.Clear();
        }
    }

    static class  objectUtil
    {
        public static Object DeepCopy(this object target)
        {
            object result;
            BinaryFormatter b = new BinaryFormatter();
            MemoryStream mem = new MemoryStream();

            try
            {
                b.Serialize(mem, target);
                mem.Position = 0;
                result = b.Deserialize(mem);
            }
            finally
            {
                mem.Close();
            }

            return result;
        }
    }
}
